from sima.sequence import Sequence, _WrapperSequence, _MotionCorrectedSequence
import numpy as np
import inspect
import sima
import sima.sequence
import cv2
import os
import warnings

from scipy.ndimage.filters import gaussian_filter
from numba import float32, int32, int16, njit

from lab3.misc.sima_compatibility import sima_compatible

class _Sequence_Suite2p(Sequence):

    """Sequence for reading binary files in the Suite2p format

    Parameters
    ----------
    ops_path : path to the ops.npy file for a suite2p dataset as a string

    """

    #TODO: multiple channels, multiple frames
    def __init__(self, ops_path):
        self._ops_path = os.path.abspath(ops_path)

        ops = np.load(ops_path, allow_pickle=True)[()]
        self._data_path = os.path.join(ops['save_path'], 'data.bin')
        self._nframes = int(ops['nframes'])
        self._Ly = ops['Ly']
        self._Lx = ops['Lx']
        self._file = open(self._data_path, 'rb')
        self._block_size = self._Ly*self._Lx*2

    def __del__(self):
        self._file.close()

    def __len__(self):
        return self._nframes

    def _get_frame(self, t):
        buff = self._file.read(self._block_size)
        data = np.frombuffer(buff, dtype=np.int16, offset=0)
        data = np.reshape(data, (1, self._Ly, self._Lx), 1)
        return data

    def _todict(self, savedir=None):
        d = {'__class__': self.__class__,
             'ops_path': self._ops_path}

        if savedir is None:
            d.update({'ops_path': os.path.abspath(self._ops_path)})
        else:
            d.update({'_abspath': os.path.abspath(self._ops_path),
                      '_relpath': os.path.relpath(self._ops_path, savedir)})
        return d



class _SpatialFilterSequence(_WrapperSequence):
    """Sequence for gaussian blurring and clipping each frame.

    Parameters
    ----------
    base : Sequence

    """

    def __init__(self, base):
        super(_SpatialFilterSequence, self).__init__(base)

    def _transform(self, frame):
        n_channels = frame.shape[-1]
        ch_frames = []
        for i in xrange(n_channels):
            ch_frame = frame[..., i]
            filtered_frame = gaussian_filter(ch_frame, sigma=3)
            clipped = np.clip(filtered_frame,
                              np.nanpercentile(filtered_frame, 50),
                              np.nanpercentile(filtered_frame, 99.5))

            ch_frames.append(clipped)

        return np.stack(ch_frames, axis=-1)

    def _get_frame(self, t):
        frame = self._base._get_frame(t)
        return np.array(map(self._transform, frame))

    def __iter__(self):
        for frame in self._base:
            yield self._transform(frame)

    @property
    def shape(self):
        return self._base.shape

    def __len__(self):
        return len(self._base)

    def _todict(self, savedir=None):
        return {
            '__class__': self.__class__,
            'base': self._base._todict(savedir),
        }


class _SplicedSequence(_WrapperSequence):
    """Sequence for splicing together nonconsecutive frames
    Parameters
    ----------
    base : Sequence
    times : list of frame indices of the base sequence
    """

    def __init__(self, base, times):
        super(_SplicedSequence, self).__init__(base)
        self._base_len = len(base)
        self._times = times

    def __iter__(self):
        try:
            for t in self._times:
                # Not sure if np.copy is needed here (see _IndexedSequence)
                yield np.copy(self._base._get_frame(t))
        except NotImplementedError:
            if self._indices[0].step < 0:
                raise NotImplementedError(
                    'Iterating backwards not supported by the base class')
            idx = 0
            for t, frame in enumerate(self._base):
                try:
                    whether_yield = t == self._times[idx]
                except IndexError:
                    raise StopIteration
                if whether_yield:
                    # Not sure if np.copy is needed here (see _IndexedSequence)
                    yield np.copy(frame)
                    idx += 1

    def _get_frame(self, t):
        return self._base._get_frame(self._times[t])

    def __len__(self):
        return len(self._times)

    def _todict(self, savedir=None):
        return {
            '__class__': self.__class__,
            'base': self._base._todict(savedir),
            'times': self._times
        }


class _MedianSequence(_WrapperSequence):
    """Sequence for applying a nan median along one of the dimensions.

    Parameters
    ----------
    base : Sequence
    axis : axis to perform nan median along

    """

    def __init__(self, base, axis=0):
        super(_MedianSequence, self).__init__(base)
        self._axis = axis
        self._shape = self._base.shape[:axis + 1] + (1,) + \
                      self._base.shape[axis + 2:]

    def _get_frame(self, t):
        frame = self._base._get_frame(t)
        return np.nanmedian(frame, axis=self._axis, keepdims=True)

    def __iter__(self):
        for frame in self._base:
            yield np.nanmedian(frame, axis=self._axis, keepdims=True)

    @property
    def shape(self):
        return self._shape

    def __len__(self):
        return len(self._base)

    def _todict(self, savedir=None):
        return {
            '__class__': self.__class__,
            'base': self._base._todict(savedir),
            'axis': self._axis}


class _NumPyFuncSequence(_WrapperSequence):
    """Sequence for applying a numypy function to each frame along one of the
    dimensions.

    Parameters
    ----------
    base : Sequence
    func : str containeng the function name to apply i.e. 'max' or 'nanmean'
    axis : axis to perform nan median along

    """

    def __init__(self, base, func, axis=0):
        super(_NumPyFuncSequence, self).__init__(base)
        self._axis = axis
        self._shape = self._base.shape[:axis + 1] + (1,) + \
            self._base.shape[axis + 2:]
        self._func = eval('np.%s' % func)
        self._func_name = func

    def _get_frame(self, t):
        frame = self._base._get_frame(t)
        return self._func(frame, axis=self._axis, keepdims=True)

    def __iter__(self):
        for frame in self._base:
            yield self._func(frame, axis=self._axis, keepdims=True)

    @property
    def shape(self):
        return self._shape

    def __len__(self):
        return len(self._base)

    def _todict(self, savedir=None):
        return {
            '__class__': self.__class__,
            'base': self._base._todict(savedir),
            'func': self._func_name,
            'axis': self._axis}


class _ClippedSequence(_WrapperSequence):
    """Sequence for clipping frame fluorescence to min and max values.

    Parameters
    ----------
    base : Sequence
    min_val : Minimum fluorescence value
    max_val: Maximum fluorescence value

    """

    def __init__(self, base, min_val, max_val):
        super(_ClippedSequence, self).__init__(base)
        # self._clip_val=99.5
        self._min_val = min_val
        self._max_val = max_val

    def _get_frame(self, t):
        frame = self._base._get_frame(t)
        return np.clip(frame, self._min_val, self._max_val)

    def __iter__(self):
        for frame in self._base:
            yield np.clip(frame, self._min_val, self._max_val)

    def __len__(self):
        return len(self._base)

    def _todict(self, savedir=None):
        return {
            '__class__': self.__class__,
            'base': self._base._todict(savedir),
            'min_val': self._min_val,
            'max_val': self._max_val
        }


class _Suite2pRigidSequence(_MotionCorrectedSequence):
    """

    Sequence wrapper for the rigid transformation from Suite2P

    Parameters
    --------------
    base : Sequence
        base sequence
    ops : dict or list[dict]
        Suite2p options dictionaries which stores the displacements

    """

    def __init__(self, base, ops=None, displacements=None):

        if ops is not None:
            displacements = self._process_ops(ops)
        super(_Suite2pRigidSequence, self).__init__(base, displacements)

    def _process_ops(self, ops):
        if len(ops) > 1:
            all_planes = []
            for plane in ops:
                all_planes.append(self._recover_displacements(plane))
            displacement = np.stack(all_planes, axis=1)
            print(displacement.shape)
            return displacement
        return self._recover_displacements(ops[0])

    def _recover_displacements(self, ops):
        _displacements = np.stack(
            [ops['yoff'], ops['xoff']]) * -1  # get the data of the x and y displacement and apply a flip
        # use the minimum displacement to translate and clip to absolute value to make sure that there are no
        # negative values, and transpose
        _displacements = np.stack([np.round(d + np.abs(np.min(d))).astype('int') for d in _displacements]).T
        # restructure the displacement shape to be according to requirements
        _displacements = _displacements.reshape(-1, 1, 2)

        return _displacements

    def _todict(self, savedir=None):
        return {
            '__class__': self.__class__,
            'base': self._base._todict(savedir),
            'displacements': self.displacements.astype('int16')
        }

#@sima_compatible
class _NoRMCorreRigidSequence(_MotionCorrectedSequence):
    """

    Sequence wrapper for the rigid transformation from NoRMCorre

    Parameters
    --------------
    base : Sequence
        base sequence
    mc : NoRMCorre motion correction object
        NoRMCorre object which stores the displacements

    """

    def __init__(self, base, mc=None, displacements=None):
        # from pudb import set_trace 
        # set_trace()
        if mc is not None:
            displacements = self._recover_displacements(mc, 
                num_planes=base.shape[1])
        super().__init__(base, displacements)

    @classmethod
    def _recover_displacements(cls, mc, num_planes=1):

        _displacements = np.array(mc.shifts_rig)

        if num_planes > 1:
            max_dz = np.max(np.abs(_displacements[:,2]))
            if max_dz > 0:
                warnings.warn(f"z-motion of up to {max_dz}px detected! "
                                f"SIMA does not handle z motion!")

            _displacements = np.repeat(_displacements[:, np.newaxis, :2], axis=1, 
                                        repeats=num_planes)
        else:
            _displacements = _displacements[:, np.newaxis, :]

        _displacements = np.round(_displacements - np.min(_displacements)).astype('int16')

        return _displacements

    def _todict(self, savedir=None):
        return {
            '__class__': self.__class__,
            'base': self._base._todict(savedir),
            'displacements': self.displacements.astype('int16')
        }


# TODO : this does not actually work
# class _BlockInterpSequence(_WrapperSequence):
#     """For storing Suite2p non-rigid motion correction results using SIMA.

#     Suite2p non-rigid correction partitions the data into XY blocks that are
#     independently motion corrected via plane translation. Here those blockwise
#     shifts are bilinearly interpolated to calculate a globally nonlinear
#     correction, applied as a piecewise affine transform.

#     TODO : Multiplane corrections. Currently we expect only one set of
#     displacements and apply these to all frames.

#     Parameters
#     ----------
#     base : Sequence
#     ops : dict
#         Suite2p options dictionary, which stores the block displacements

#     """

#     def __init__(self, base, ops):
#         super(_BlockInterpSequence, self).__init__(base)
#         self._ops = ops

#         # get block information
#         nblocks = ops['nblocks']
#         yblock = ops['yblock'][::-1]
#         xblock = ops['xblock'][::-1]

#         # reshape and store displacements
#         yshifts = (ops['yoff1'] * -1).astype('float32').reshape(
#             -1, nblocks[0], nblocks[1])
#         xshifts = (ops['xoff1'] * -1).astype('float32').reshape(
#             -1, nblocks[0], nblocks[1])
#         self.displacements = np.stack([yshifts, xshifts], axis=-1)

#         num_rows, num_columns = self._base.shape[2:4]

#         # save mesh grids for later interpolations
#         yb = np.array(yblock[::nblocks[1]]).mean(axis=1).astype(np.float32)
#         iy = np.arange(0, num_rows, 1, np.float32)
#         iy = np.interp(iy, yb, np.arange(0, yb.size, 1, int)).astype(np.float32)

#         xb = np.array(xblock[:nblocks[1]]).mean(axis=1).astype(np.float32)
#         ix = np.arange(0, num_columns, 1, np.float32)
#         ix = np.interp(ix, xb, np.arange(0, xb.size, 1, int)).astype(np.float32)

#         self.imshx, self.imshy = np.meshgrid(ix, iy)
#         self.mshx, self.mshy = np.meshgrid(
#             np.arange(0, num_columns, 1, np.float32),
#             np.arange(0, num_rows, 1, np.float32))

#     def __len__(self):
#         return len(self._base)

#     def _transform(self, frame, t):
#         # upsample block shifts
#         yup = bilinear_transform(
#             self.displacements[t][..., 0], self.imshy.copy(),
#             self.imshx.copy()).astype('float32')
#         xup = bilinear_transform(
#             self.displacements[t][..., 1], self.imshy.copy(),
#             self.imshx.copy()).astype('float32')

#         out = np.nan * np.ones(frame.shape)
#         for pidx, plane in enumerate(frame):
#             for cidx in range(plane.shape[-1]):
#                 out[pidx, :, :, cidx] = bilinear_transform(
#                     plane[..., cidx].astype('int16'),
#                     self.mshy + yup, self.mshx + xup)

#         return out
    
#     def __iter__(self):
#         for frame in self._base:
#             yield self._transform(frame)

#     def _get_frame(self, t):
#         return self._transform(self._base._get_frame(t), t)  # add args
    
#     def _todict(self, savedir=None):
#         return {
#             '__class__': self.__class__,
#             'base': self._base._todict(savedir),
#             'ops': self._ops
#         }


class _InterpolateGapFilledSequence(_WrapperSequence):
    """Sequence for doing random stuff.

    Parameters
    ----------
    base : Sequence
    kernel_size : Size of 2D gaussian kernel to use for interpolating nans.

    """

    def __init__(self, base, kernel_size):
        super(_InterpolateGapFilledSequence, self).__init__(base)

        from astropy.convolution import Gaussian2DKernel
        from astropy.convolution import interpolate_replace_nans

        self._kernel_size = kernel_size
        self._shape = self._base.shape
        self._interp_fn = interpolate_replace_nans
        self._kernel = Gaussian2DKernel(self._kernel_size)
        self._cutoff = np.prod(self._shape[1:])/3

        seq_dict = base._todict()
        self._unmasked_base = seq_dict.pop('__class__')._from_dict(seq_dict)
        seq_copy = self._unmasked_base
        self._masked = True
        while True:
            if isinstance(seq_copy._base, sima.sequence._MaskedSequence):
                    break

            try:
                seq_copy._base._base
                seq_copy = seq_copy._base
            except:
                self._masked = False
                break

        if self._masked:
            seq_copy._base = seq_copy._base._base
            self.__iter__ = self._iter_masked
            self._get_frame = self._get_frame_masked
        else:
            self._unmasked_base = None

    def _transform_masked(self, frame_masked, frame_unmasked):
        nan_locations = np.where(np.isnan(frame_masked))[0].shape[0]
        if nan_locations > self._cutoff:
            return frame_masked

        frame_mask = np.where(
            np.logical_and(np.isnan(frame_masked), np.isfinite(frame_unmasked)))

        _frame = np.empty(frame_unmasked.shape)
        for p, plane in enumerate(frame_unmasked):
            for c, channel in enumerate(np.rollaxis(plane, 2, 0)):
                _frame[p, :, :, c] = self._interp_fn(channel, self._kernel)
        _frame[frame_mask] = np.nan
        return _frame

    def _get_frame_masked(self, t):
        frame_masked = self._base._get_frame(t)
        frame_unmasked = self._unmasked_base._get_frame(t)
        return self._transform_masked(frame_masked, frame_unmasked)

    def _iter_masked(self):
        for frame_masked, frame_unmasked in zip(self._base,
                                                self._unmasked_base):
            yield self._transform_masked(frame_masked, frame_unmasked)

    def _transform(self, frame):
        _frame = np.empty(frame.shape)
        for p, plane in enumerate(frame):
            for c, channel in enumerate(np.rollaxis(plane, 2, 0)):
                _frame[p, :, :, c] = self._interp_fn(channel, self._kernel)
        return _frame

    def _get_frame(self, t):
        frame = self._base._get_frame(t)
        return self._transform(frame)

    @property
    def shape(self):
        return self._shape

    def __len__(self):
        return len(self._base)

    def _todict(self, savedir=None):
        return {
            '__class__': self.__class__,
            'base': self._base._todict(savedir),
            'kernel_size': self._kernel_size
        }


@njit(['(int16[:, :],float32[:,:], float32[:,:])',
       '(float32[:, :],float32[:,:], float32[:,:])'])
def bilinear_transform(I, yc, xc):
    """Doc string"""

    I_transform = np.zeros((yc.shape))

    num_rows, num_columns = I.shape
    yc_floor, xc_floor = [c.copy().astype(np.int32) for c in [yc, xc]]
    yc -= yc_floor
    xc -= xc_floor

    # calculate value of each pixel
    for i in range(yc_floor.shape[0]):
        for j in range(yc_floor.shape[1]):
            yf = min(num_rows - 1, max(0, yc_floor[i, j]))
            xf = min(num_columns - 1, max(0, xc_floor[i, j]))
            yf1 = min(num_rows - 1, yf + 1)
            xf1 = min(num_columns - 1, xf + 1)
            I_transform[i, j] = \
                np.float32(I[yf, xf]) * (1 - yc[i, j]) * (1 - xc[i, j]) \
                + np.float32(I[yf, xf1]) * (1 - yc[i, j]) * xc[i, j] \
                + np.float32(I[yf1, xf]) * yc[i, j] * (1 - xc[i, j]) \
                + np.float32(I[yf1, xf1]) * yc[i, j] * xc[i, j]

    return I_transform


class _AffineRegisteredSequence(_WrapperSequence):
    """Wraps a SIMA sequence to apply affine warp matrix. Best on 
    already MC-wrapped sequences that are aligned to themselves such 
    that a single warp matrix is appropriate for the whole sequence.

    Parameters
    __________
    base : SIMA sequence, typically already MC'd
    warpMat : warp matrix to transform sequence into alignment
        with the template image used for cross-registration. Warp
        matrices for all wrapped sequences within an affine cross-
        registered dataset will align their respective sequences to
        the same shape (to whatever extent possible).
    ECC : enhanced correlation coefficient (calculated by 
        cv2.findTransformECC) which reflects goodness of fit between
        transformed image and the reference it was mapped to. value of
        -1 indicates the ML algorithm (maximizing ECC as obj func) 
        failed to converge.
    """

    def __init__(self, base, warpMat, ECC, refDims):
        super(_AffineRegisteredSequence, self).__init__(base)

        self.warpMat = warpMat
        self.ECC = float(ECC)
        self.refDims = refDims

        if ECC == -1:
            raise ValueError('warpMat == identity matrix. No transform will occur.')


    @property
    def _frame_shape(self):
        return self._frame_shape #inherited from MC wrapper

    def __len__(self):
        return len(self._base)


    def __iter__(self):
        for frame in self._base:
            yield self._align(frame)


    #return 4D aligned frame (zyxc)
    def _align(self, frame):
        assert len(frame.shape) == 4 #zyxc
        nPlanes = frame.shape[0]
        nChannels = frame.shape[3]
        alignedFrame = np.zeros((
            nPlanes, self.refDims[0], self.refDims[1], nChannels))
        for p in np.arange(nPlanes):
            for c in np.arange(nChannels):
                currXY = frame[p, :, :, c]
                alignedXY = cv2.warpAffine(currXY, self.warpMat, 
                    (self.refDims[1], self.refDims[0]), 
                    flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP)
                alignedFrame[p, :, :, c] = alignedXY

        return alignedFrame

    def _get_frame(self, t):
        return self._align(self._base._get_frame(t))

    def _todict(self, savedir=None):
        return{
        '__class__': self.__class__,
        'base': self._base._todict(savedir),
        'warpMat': self.warpMat,
        'ECC': self.ECC,
        'refDims': self.refDims
        }


sima.sequence.__dict__.update(
    {k: v for k, v in locals().items() if
        inspect.isclass(v) and (issubclass(v, sima.sequence._WrapperSequence) or
        issubclass(v, sima.sequence.Sequence))})



import numpy as np 
import scipy.stats as stats 
import statsmodels
from statsmodels.tsa.statespace import sarimax
from statsmodels.regression.linear_model import yule_walker

from abc import abstractmethod
import pandas as pd 

import lab3

from lab3.core.filters import IndexIdentity, ColumnIdentity

from lab3.analysis.base import SignalAnalysis
from lab3.analysis.basic.psth import Responsiveness
from lab3.analysis.spatial.abstract_spatial_tuning import SpatialTuning

from lab3.core.helpers import save_load

class Feature(type):
    def __new__(cls, func, name=None):
        if name is None:
            name = func.__name__.capitalize()
        
        return type(name, (FeatureExtractor,), 
                    {
                        'func': staticmethod(func), 
                     '__doc__': func.__doc__,
                     '__module__': __name__ #lab #lab3.analysis.basic.features
                    }
                   )
    
class SpatialFeature(type):
    def __new__(cls, func, name=None):
        if name is None:
            name = func.__name__.capitalize()
        
        return type(name, (SpatialFeatureExtractor,), 
                    {
                        'func': staticmethod(func), 
                     '__doc__': func.__doc__,
                     '__module__':__name__ #lab# lab3.analysis.basic.features
                    }
                   )

class ResponsivenessFeature(type):
    def __new__(cls, event_spec, name=None):
        if name is None:
            name = event_spec['event_type'].capitalize()
        
        return type(name, (ResponsivenessFeatureExtractor,), 
                    {
                        'events': event_spec, 
                        '__module__': __name__ #lab# lab3.analysis.basic.features
                    }
                   )
    
class FeatureExtractor(SignalAnalysis):
    """Generic class for feature extraction, to feed to machine learning
    """
    constructor_ids = ['roi_filter', 'time_filter']
    def to_columns(self):
        return pd.Index([self.name])
        
    def __init__(self, roi_filter=IndexIdentity(), time_filter=ColumnIdentity()):
        self.name = self.__class__.__name__
        self.roi_filter = roi_filter
        self.time_filter = time_filter
        
    @save_load
    def apply_to(self, expt, signal_spec):
        signals = self._load_signals(expt, signal_spec)
        signals = signals.loc[self.roi_filter.bind_to(expt), 
                              self.time_filter.bind_to(expt)]
        
        return self.calculate(signals)
    
    def calculate(self, signals):
        result = signals.apply(self.func, axis=1)
        try:
            result = result.to_frame()
        except AttributeError:
            pass
        result.columns = self.columns
        return result
    
    
class SpatialFeatureExtractor(SpatialTuning, FeatureExtractor):
    
    constructor_ids = SpatialTuning.constructor_ids + FeatureExtractor.constructor_ids
    
    def to_columns(self):
        return pd.Index([self.name])
    
    def __init__(self, nbins: int = 100, sigma: float = 3, 
                 normalize: bool = True, roi_filter=IndexIdentity(), 
                 time_filter=ColumnIdentity()):
        self.name = self.__class__.__name__
        self.roi_filter = roi_filter
        self.time_filter = time_filter
        self.nbins = nbins 
        self.bins = None
        self.sigma = sigma 
        self.normalize = normalize
        self.roi_filter = roi_filter
        self.time_filter = time_filter
        self.spatial_tuning_strategy = SpatialTuning(nbins=nbins, bins=None, sigma=sigma, 
                                                        normalize=normalize)
        
    def apply_to(self, expt, signal_spec):
        tuning_curves = self.spatial_tuning_strategy.apply_to(
                                        expt, signal_spec=signal_spec, 
                                         roi_filter=self.roi_filter, 
                                         time_filter=self.time_filter)
        tuning_curves = tuning_curves.groupby('roi_label').apply(np.mean)
        tuning_curves = tuning_curves.fillna(0) # Set nans to 0
        return self.calculate(tuning_curves)
    
    def calculate(self, tuning_curves):
        result = tuning_curves.apply(self.func, axis=1).to_frame()
        result.columns = self.columns
        return result

class ResponsivenessFeatureExtractor(FeatureExtractor, Responsiveness):
    constructor_ids = Responsiveness.constructor_ids + FeatureExtractor.constructor_ids + ['contrast']
    
    # def to_columns(self):
    #   return pd.Index([self.name])

    def __init__(self, pre: float = 3., post: float = 3., dt: float = 0.03, contrast: bool = False,
                 roi_filter=IndexIdentity(), time_filter=ColumnIdentity()):
        self.name = self.__class__.__name__
        self.response_strategy = Responsiveness(pre=pre, post=post, dt=dt)
        self.pre = pre 
        self.post = post 
        self.dt = dt 
        self.roi_filter = roi_filter
        self.time_filter = time_filter
        self.contrast = contrast
    
    def apply_to(self, expt, signal_spec):
        pre_post = self.response_strategy.apply_to(expt, events=self.events, signal_spec=signal_spec, 
                                    roi_filter=self.roi_filter, time_filter=self.time_filter)

        groupby = [label for label in pre_post.index.names if 'Event' not in label]
        pre_post = pre_post.groupby(groupby).mean()

        responsiveness = pre_post['post'] - pre_post['pre']

        if self.contrast:
            responsiveness /= pre_post['post'] + pre_post['pre']

        responsiveness = responsiveness.to_frame()
        responsiveness.columns = self.columns

        return responsiveness

class Moment(FeatureExtractor):
    constructor_ids = ['moment', 'roi_filter', 'time_filter']
    
    def to_columns(self):
        return pd.Index([f"Moment[{self.moment}]"])
    
    def __init__(self, moment, roi_filter=IndexIdentity(), time_filter=ColumnIdentity()):
        self.moment = moment
        super().__init__(roi_filter=roi_filter, time_filter=time_filter)
        
    def func(self, x):
        return stats.moment(x, moment=self.moment, nan_policy='omit')
        
class Cumulant(FeatureExtractor):
    constructor_ids = ['cumulant', 'roi_filter', 'time_filter']
    
    def to_columns(self):
        return pd.Index([f"Cumulant[{self.cumulant}]"])
    
    def __init__(self, cumulant, roi_filter=IndexIdentity(), time_filter=ColumnIdentity()):
        self.cumulant = cumulant
        assert 1<= cumulant <= 4,\
            "Only cumulants up to 4 are implemented!"
        super().__init__(roi_filter=roi_filter, time_filter=time_filter)
        
    def func(self, x):
        return stats.kstat(x, n=self.cumulant)

class ARCoef(FeatureExtractor):
    constructor_ids = ['p', 'method', 'roi_filter', 'time_filter']
    
    def to_columns(self):
        try:
            p = self.p[0]
            d = self.p[1]
        except TypeError:
            p = self.p
            d = 0
        
        p_features = [f'AR[{i}]' for i in range(p)]
        d_features = [f'MA[{i}]' for i in range(d)]    
        
        return pd.Index(p_features + d_features)
    
    def __init__(self, p, method='yule_walker', roi_filter=IndexIdentity(), 
                 time_filter=ColumnIdentity(), **ar_kwargs):
        """Fits an AR(p) model to the signals and extracts the AR coefficients
        as features.
        
        Inputs
        -------
        p : int or tuple of int
            The order of the autoregressive process
        method : str, 'yule_walker' or 'sarimax'
            If you don't know what these are, use 'yule_walker'
        roi_filter : ROIFilter
        time_filter : TimeFilter
        **ar_kwargs : dict
            Keyword arguments passed to either yule_walker or SARIMAX 
        """
        self.p = p
        self.ar_kwargs = ar_kwargs
        self.method = method
        
        if self.method == 'yule_walker':
            self.func = self.yule_walker
        elif self.method == 'sarimax':
            self.func = self.sarimax
        
        super().__init__(roi_filter=roi_filter, time_filter=time_filter)
    
    def yule_walker(self, x):
        return pd.Series(yule_walker(x, order=self.p, **self.ar_kwargs)[0])
    
    def sarimax(self, x):
        mod = sarimax.SARIMAX(x, order=self.p, **self.ar_kwargs)
        res = mod.fit()
        return np.concatenate([res.arparams, res.maparams])
    
# NB: It's important here that the class name and the `name` attribute MATCH!
# Otherwise pickle-dependent operations (including parallelization) will break

Mean = Feature(np.nanmean, name="Mean")
StandardDeviation = Feature(np.nanstd, name="StandardDeviation")
Variance = Feature(np.nanvar, name="Variance")
Max = Feature(np.nanmax, name="Max")
Min = Feature(np.nanmin, name="Min")
Median = Feature(np.nanmedian, name="Median")
#Mode = Feature(stats.mode)
Skew = Feature(stats.skew)
Kurtosis = Feature(stats.kurtosis)
IQR = Feature(stats.iqr)
GMean = Feature(stats.gmean)
HMean = Feature(stats.hmean)

Entropy = SpatialFeature(stats.entropy)
Peak = SpatialFeature(np.argmax, name="peak")
MAD = SpatialFeature(stats.median_absolute_deviation, name="MAD")

RunningStart = ResponsivenessFeature({'signal_type': 'behavior', 
                                    'event_type': 'running', 
                                    'fmt': 'onset'}, name='RunningStart')
RunningStop = ResponsivenessFeature({'signal_type': 'behavior', 
                                    'event_type': 'running', 
                                    'fmt': 'onset'}, name='RunningStop')
Reward = ResponsivenessFeature({'signal_type': 'behavior', 
                                'event_type': 'reward', 
                                'fmt': 'onset'})
Water = ResponsivenessFeature({'signal_type': 'behavior', 
                                'event_type': 'water', 
                                'fmt': 'onset'})
LED_on = ResponsivenessFeature({'signal_type': 'behavior', 
                                'event_type': 'led_context_pin', 
                                'fmt': 'onset'}, name='LED_on')
LED_off = ResponsivenessFeature({'signal_type': 'behavior', 
                                'event_type': 'led_context_pin', 
                                'fmt': 'offset'}, name='LED_off')
Licking = ResponsivenessFeature({'signal_type': 'behavior', 
                                    'event_type': 'licking', 
                                    'fmt': 'onset'})



from __future__ import annotations

import numpy as np
from scipy.ndimage import gaussian_filter1d
import scipy.stats as stats 
import pandas as pd
import itertools as it

from typing import TYPE_CHECKING

from lab3.core import Analysis
from lab3.core.filters import IndexIdentity, ColumnIdentity
from lab3.core.helpers import save_load
from lab3.analysis.base import SignalAnalysis
from lab3.event import Event, EventGroup, LapEvent, LapGroup

from lab3.core.helpers import save_load

if TYPE_CHECKING:
    from lab3.experiment import ImagingExperiment

class _SingleLapSpatialTuning(SignalAnalysis):
    name = 'SpatialTuning'
    constructor_ids = ['nbins', 'bins', 'sigma']
    def to_columns(self):
        return pd.Index(self.get_bins(nbins=self.nbins, bins=self.bins), name='position')
    
    def __init__(self, nbins: int = 20, bins=None, sigma: float = 3, 
                normalize: bool = True, save_results=True, load_saved=True, verbose=False):
        self.bins = self.get_bins(nbins, bins)
        self.nbins = len(self.bins)
        self.sigma = sigma
        self.normalize = normalize
        self.save_results = save_results
        self.load_saved = load_saved
        self.verbose = verbose
        self.loaded = False
        
    def apply_to(self, lap : LapEvent, signals, positions):
        onset = int(lap.onset)
        offset = int(lap.offset)
        return self._calculate(signals, positions, onset, offset)
    
    def _calculate(self, signals, positions, onset, offset):
        lap_signals = signals.iloc[:, onset:offset]
        lap_positions = positions[onset:offset]

        lap_signals.columns = self._to_bin(lap_positions)
        if self.normalize:
            tuning_curves = lap_signals.T.groupby('position').mean().T
        else:
            tuning_curves = lap_signals.T.groupby('position').sum().T
        
        if self.sigma is not None:
            tuning_curves = self._smooth(tuning_curves)
        
        return tuning_curves
    
    def _smooth(self, tuning_curves):        
        # This is an inplace operation
        gaussian_filter1d(tuning_curves, self.sigma, axis=1, 
        					output=tuning_curves, mode='wrap')
        return tuning_curves
    
    def _to_bin(self, x):
        return self.columns[np.searchsorted(self.bins, x, side='right') - 1]
    
    @classmethod
    def get_bins(cls, nbins=None, bins=None):
        if nbins is not None:
            bins = np.linspace(0,1,nbins+1)[:-1]
        elif nbins is not None:
            bins = np.sort(bins)
            nbins = len(bins)
        else:
            raise ValueError("One of `bins` and `nbins` must be specified!")
            
        return bins
    
class SpatialTuning(_SingleLapSpatialTuning):
    """A simple spatial tuning strategy. For each neuron, we compute its
    average activity across samples/laps for each position. The tuning curves
    are optionally smoothed with a Gaussian kernel. Since we are computing
    the average activity across samples for a given position, this implicitly
    normalizes for occupancy.

    Initialization parameters
    ----------
    nbins : int
        Number of bins to discretize the track
    sigma : float, optional
        Standard deviation of the gaussian smoothing kernel applied to the
        tuning curves. By default, no smoothing is applied.

    Examples
    --------
    >>> # Compute the tuning curves
    >>> strategy = SpatialTuning(sigma=2)
    >>> signal_spec = {"signal_type": "dfof", "label": "suite2p"}
    >>> tuning_curves = strategy.apply_to(
    >>>     expt, signal_spec=signal_spec)
    """

    @save_load
    def apply_to(self, expt, signal_spec, roi_filter=IndexIdentity(), 
                    time_filter=ColumnIdentity(), verbose=False, behavior_spec={}):
        signals = self._load_signals(expt, signal_spec)
        signals = signals.loc[roi_filter.bind_to(expt)]

        # Because of laps and positions, these cannot simply be 
        # stripped out
        # TODO: Is replacing with nan ALWAYS preferable?
        signals.loc[:, ~(time_filter.bind_to(expt))] = np.nan

        laps = LapGroup.from_expt(expt)
        positions = expt.format_behavior_data(**behavior_spec)['treadmillPosition']

        return self.calculate(signals, laps, positions)
 
    def calculate(self, signals, laps, positions):
        return laps.apply(super(), signals=signals, positions=positions)


class EventSpatialTuning(SpatialTuning):
    
    def apply_to(self, expt, event_spec, **kwargs):
        events = expt.events(**event_spec)
        index  = expt.signals().index
        
        positions = expt.format_behavior_data()['treadmillPosition']
        imaging_times = expt.imaging_times()
        
        return self.calculate(events, positions, imaging_times, index)
    
    def calculate(self, events, positions, imaging_times, index):
        def _to_position(onset):
            idx = np.searchsorted(imaging_times, onset) - 1
            return positions[idx]
                
        results = pd.DataFrame(np.zeros((len(index),self.nbins)), 
                               index=index, 
                               columns=self.columns)
        
        bins = events.onset.map(lambda x: self._to_bin(_to_position(x)))
        counts = bins.value_counts()
        results.loc[:, counts.index] += counts
        
        return results

class Occupancy(SpatialTuning):
    """Calculates the amount of time spent in each bin
    """
    name = 'Occupancy'
    def apply_to(self, expt, *args, **kwargs):
        occupancy_normalize = self.normalize
        self.normalize = False 

        result = super().apply_to(expt, *args, **kwargs)
        result.index = result.index.droplevel(None)

        if occupancy_normalize:
            result /= result.sum()

        return result

    def _load_signals(self, expt, signal_spec):
        signals = super()._load_signals(expt, signal_spec)
        signals = pd.DataFrame(np.ones((1,signals.shape[1])), 
                               columns=signals.columns)
        return signals

class PopulationVectorCorrelations(SpatialTuning):
    constructor_ids = ['nbins', 'bins', 'sigma', 'correlation_function']
    
    def to_columns(self):
        return pd.Index(['correlation', 'pval'])
    
    def __init__(self, correlation_function=stats.pearsonr, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.spatial_tuning = SpatialTuning(*args, **kwargs)
        self.correlation_function = correlation_function
        
    def apply_to(self, expt : ImagingExperiment, signal_spec, 
                     roi_filter=IndexIdentity(), 
                     time_filter=ColumnIdentity(), time_filter2=None, 
                 **corr_kwargs):
        tuning_curves = self.spatial_tuning.apply_to(expt, signal_spec=signal_spec, 
                                                     roi_filter=roi_filter, 
                                                     time_filter=time_filter)
        tuning_curves = tuning_curves.groupby('roi_label').agg('mean')
        
        if time_filter2 is not None:
            tuning_curves2 = self.spatial_tuning.apply_to(expt, signal_spec=signal_spec, 
                                                         roi_filter=roi_filter, 
                                                         time_filter=time_filter2)
            tuning_curves2 = tuning_curves2.groupby('roi_label').agg('mean')            
        else:
            tuning_curves2 = None
        
        return self.calculate(tuning_curves, tuning_curves2, **corr_kwargs)           
        
    def calculate(self, tuning_curves, tuning_curves2=None, **corr_kwargs):

        if tuning_curves2 is None:
            tuning_curves2 = tuning_curves
        
        def corr(series):
            x = tuning_curves.loc[:, series.bin_1] 
            y = tuning_curves2.loc[:, series.bin_2]

            is_na = pd.isnull(x) + pd.isnull(y)

            try:
                return self.correlation_function(x.loc[~is_na], 
                                             y.loc[~is_na], 
                                             **corr_kwargs)
            except:
                return np.nan
        
        pairs = pd.DataFrame(it.product(tuning_curves.columns, 
                                        tuning_curves2.columns), 
                             columns=['bin_1', 'bin_2'])
        
        correlation_tuples = pairs.apply(corr, axis=1)
        
        result = pd.DataFrame([tup for tup in correlation_tuples],
                              index=pd.MultiIndex.from_frame(pairs),
                              columns=self.columns)
        
        return result
    
class TuningCurveCorrelations(SpatialTuning):
    constructor_ids = ['nbins', 'bins', 'sigma', 'correlation_function']
    
    def to_columns(self):
        return pd.Index(['correlation', 'pval'])
    
    def __init__(self, correlation_function=stats.pearsonr, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.spatial_tuning = SpatialTuning(*args, **kwargs)
        self.correlation_function = correlation_function
        
    def apply_to(self, expt : ImagingExperiment, signal_spec, 
                     roi_filter=IndexIdentity(), 
                     time_filter=ColumnIdentity(), time_filter2=None, 
                     **corr_kwargs):
        tuning_curves = self.spatial_tuning.apply_to(expt, signal_spec=signal_spec, 
                                                     roi_filter=roi_filter, 
                                                     time_filter=time_filter)
        tuning_curves = tuning_curves.groupby('roi_label').agg('mean')
        
        if time_filter2 is not None:
            tuning_curves2 = self.spatial_tuning.apply_to(expt, signal_spec=signal_spec, 
                                                         roi_filter=roi_filter, 
                                                         time_filter=time_filter2)
            tuning_curves2 = tuning_curves2.groupby('roi_label').agg('mean')            
        else:
            tuning_curves2 = None
        
        return self.calculate(tuning_curves, tuning_curves2, **corr_kwargs)        
        
    def calculate(self, tuning_curves, tuning_curves2=None, **corr_kwargs):
        
        if tuning_curves2 is None:
            tuning_curves2 = tuning_curves
        
        def corr(series):
            x = tuning_curves.loc[series.cell_1] 
            y = tuning_curves2.loc[series.cell_2]
            
            is_na = pd.isnull(x) + pd.isnull(y)
            
            return self.correlation_function(x.loc[~is_na], 
                                             y.loc[~is_na], 
                                             **corr_kwargs)
    def calculate(self, tuning_curves, tuning_curves2=None, **corr_kwargs):
        
        if tuning_curves2 is None:
            tuning_curves2 = tuning_curves
        
        def corr(series):
            x = tuning_curves.loc[series.cell_1] 
            y = tuning_curves2.loc[series.cell_2]
            
            is_na = pd.isnull(x) + pd.isnull(y)
            
            return self.correlation_function(x.loc[~is_na], 
                                             y.loc[~is_na], 
                                             **corr_kwargs)
        
        pairs = pd.DataFrame(it.product(tuning_curves.index, 
                                        tuning_curves2.index), 
                             columns=['cell_1', 'cell_2'])
        
        correlation_tuples = pairs.apply(corr, axis=1)
        
        result = pd.DataFrame([tup for tup in correlation_tuples],
                              index=pd.MultiIndex.from_frame(pairs),
                              columns=self.columns)
        
        return result



"""Filter functions for signals. Functions in this module should take as their
first argument an array-like object of size (n_ROIs, n_timepoints), and should
return results in an identical format. If the input is a DataFrame, this
should be preserved (along with its index)."""


import pandas as pd


def maxmin_filter(signal, window=300, sigma=5, min_periods=0.2):
    """Calculate baseline as the rolling maximum of the rolling minimum of the
    smoothed trace

    Parameters
    ----------
    signal : array, size (n_ROIs, n_timepoints)
    window : int
        Optional, size of the rolling window for max/min/smoothing
    sigma : int
        Standard deviation of the gaussian smoothing kernel
    min_periods : float [0, 1]
        Optional, percentage of values in each window that must be non-NaN to
        return a non-NaN value in the baseline
    """

    kwargs = {'window': window, 'min_periods': int(window * min_periods),
              'center': True, 'axis': 1}

    smooth_signal = pd.DataFrame(signal).rolling(
        win_type='gaussian', **kwargs).mean(std=sigma)

    return smooth_signal.rolling(**kwargs).min().rolling(**kwargs).max()


def smooth_quantile_filter(signal, quantile=0.08, window=300, sigma=5,
                           min_periods=0.2):

    kwargs = {'window': window, 'min_periods': int(window * min_periods),
              'center': True, 'axis': 1}

    smooth_signal = pd.DataFrame(signal).rolling(
        win_type='gaussian', **kwargs).mean(std='sigma')

    return smooth_signal.rolling(**kwargs).quantile(quantile)



import abc
import pickle as pkl 
import pandas as pd
import numpy as np

from sklearn.preprocessing import scale

from lab3.core.filters import ROIFilter, ColumnIdentity
from lab3.analysis.spatial.abstract_place_fields import PlaceFields
from lab3.analysis.basic.psth import Responsiveness

from lab3.experiment.base import ImagingExperiment

class ClassifierROIFilter(ROIFilter):
    
    constructor_ids = ['feature_extractor', 'trained_classifier', 'signal_spec']
    
    @classmethod
    def load_pkl(cls, path_to_pkl, signal_spec):
        with open(path_to_pkl, "rb") as f:
            pickled_filter = pkl.load(f)
        
        return cls(feature_extractor=pickled_filter['feature_extractor'], 
                   trained_classifier=pickled_filter['trained_classifier'], 
                   signal_spec=signal_spec)
    
    def bind_to(self, expt):
        X = self.feature_extractor.apply_to(expt, signal_spec=self.signal_spec)
        X = X.fillna(0)
        X = pd.DataFrame(scale(X), index=X.index, columns=X.columns)
        y = pd.Series(self.trained_classifier.predict(X), index=X.index)
        
        return self.bind(y[y>0].index)

class ResponsivenessFilter(ROIFilter, Responsiveness):
    constructor_ids = Responsiveness.constructor_ids + ['threshold', 'signal_spec', 'event_spec']
    
    def __init__(self, signal_spec, event_spec, threshold=0.5, **kwargs):
        self.signal_spec = signal_spec
        self.event_spec = event_spec
        self.threshold = threshold
        
        super().__init__(**kwargs)

    @property
    def name(self):
        return f"Is{self.event_spec['event_type'].capitalize()}Responsive"

        
    def _repr_html_(self):
        return ROIFilter.__repr__(self)
    
    def bind_to(self, expt : ImagingExperiment):
        responses = super().apply_to(expt, signal_spec=self.signal_spec, 
                                     events=self.event_spec)
        responses = responses.groupby('roi_label').agg('mean')
        scalar_responses = responses['post'] - responses['pre']
        is_responsive = scalar_responses[scalar_responses > self.threshold].index
        return self.bind(is_responsive)

class IsPlaceCell(ROIFilter, PlaceFields):
    constructor_ids = PlaceFields.constructor_ids + ['signal_spec', 'time_filter']

    def __init__(self, *args, signal_spec={}, time_filter=ColumnIdentity, **kwargs):
        self.signal_spec = signal_spec
        self.time_filter = time_filter
        super().__init__(*args, **kwargs)

    def bind_to(self, expt):
        pfs = super().apply_to(expt, signal_spec=self.signal_spec, 
                                time_filter=self.time_filter)
        return self.bind(index=pfs.level['roi_label'].unique())

class IsInterneuron(ClassifierROIFilter):
    def __init__(self, signal_spec={"signal_type": "dfof"}):
        with open('/data/Zhenrui/classifiers/is_interneuron.pkl', 'rb') as f:
            pickled_filter = pkl.load(f)
        
        super().__init__(feature_extractor=pickled_filter['feature_extractor'], 
                   trained_classifier=pickled_filter['trained_classifier'], 
                   signal_spec=signal_spec)

class PlaceFieldLocationFilter(IsPlaceCell):
    """Filter cells by the location of their place fields
    
    Examples
    --------
    Select cells with place fields within the zone of interest.
    >>> place_field_starts = PlaceFieldStart(signal_spec=signal_spec)
    >>> place_field_stops = PlaceFieldStop(signal_spec=signal_spec)
    >>> zone_of_interest = (0.2, 0.4)
    >>> pcs_in_zone = (0.2 < place_field_starts) & (place_field_stops < 0.4)
    >>> 
    
    Important: Remember the belt is circular! Obviously there does not exist 
    a well-ordering on the circle. It is up to you to handle behavior at the reset.
    """
    constructor_ids = IsPlaceCell.constructor_ids + ['lower', 'upper']
    def __init__(self, *args, lower=0, upper=1, **kwargs):
        self.lower = lower
        self.upper = upper
        super().__init__(*args, **kwargs)
    
    @property
    def name(self):
        if self.lower == 0:
            return f"({self.__class__.__name__} < {self.upper})"
        elif self.upper == 1:
            return f"({self.lower} < {self.__class__.__name__})"
        else:
            return f"({self.lower} < {self.__class__.__name__} < {self.upper})"
        
    def __gt__(self, position):
        cons = self.constructors()
        cons['lower'] = position
        return self.__class__(**cons)
    
    def __lt__(self, position):
        cons = self.constructors()
        cons['upper'] = position
        return self.__class__(**cons)
    
    def __ge__(self, position):
        return ~(self < position)
    
    def __le__(self, position):
        return ~(self > position)
    
    def _repr_html_(self):
        raise NotImplementedError
    
    def pos2bin(self, position):
        return np.searchsorted(np.linspace(0,1,self.num_bins), position, side='right') - 1
    
    @abc.abstractmethod
    def bind_to(self, expt):
        """"""

class PlaceFieldStart(PlaceFieldLocationFilter):
    def bind_to(self, expt):
        pfs = super().apply_to(expt, signal_spec=self.signal_spec)
        
        lower_bin = self.pos2bin(self.lower)
        upper_bin = self.pos2bin(self.upper)
        
        pfs = pfs.loc[(lower_bin < pfs.field_start) & (pfs.field_start < upper_bin)]
        
        return self.bind(index=pfs.level['roi_label'].unique())
    
class PlaceFieldStop(PlaceFieldLocationFilter):
    def bind_to(self, expt):
        pfs = super().apply_to(expt, signal_spec=self.signal_spec)
        
        lower_bin = self.pos2bin(self.lower)
        upper_bin = self.pos2bin(self.upper)
        
        pfs = pfs.loc[(lower_bin < pfs.field_stop) & (pfs.field_stop < upper_bin)]
        
        return self.bind(index=pfs.level['roi_label'].unique())



import numpy as np 
import pandas as pd 

from lab3.metadata import DefaultMetadata

class SignalGetter:
    def __call__(self, expt, **kwargs):
        raise NotImplementedError

class Event2Signals(SignalGetter):
    def __init__(self, event_spec, signal_type, 
            groupby='roi_label', metadata=DefaultMetadata): 
        self.event_spec = event_spec
        self.groupby = groupby
        self.signal_type = signal_type
        self.metadata = metadata
    
    def __call__(self, expt, **kwargs):
        raise NotImplementedError
    
class EventOnsets2Signals(Event2Signals):
    def __call__(self, expt, **kwargs):
        event_df = expt.events(**self.event_spec).dataframe
        
        signals = expt.signals(signal_type=self.signal_type, 
                    channel=self.event_spec['channel'], metadata=self.metadata)
        signals.loc[:,:] = 0
        
        for roi_label, events in event_df.groupby(self.groupby):
            onsets = events.onset.astype(int)
            idx = signals.level[self.groupby].get_loc(roi_label)
            signals.iloc[idx, onsets] = 1.
            
        return signals

class TimeOnsets2Signals(EventOnsets2Signals):
    def __call__(self, expt, **kwargs):
        imaging_times = expt.imaging_times()
        event_df = expt.events(**self.event_spec).dataframe
        
        signals = expt.signals(signal_type=self.signal_type, 
                    channel=self.event_spec['channel'], metadata=self.metadata)
        signals.loc[:,:] = 0
        
        for roi_label, events in event_df.groupby(self.groupby):
            onsets = events.onset
            frame_idxes = np.searchsorted(imaging_times, onsets) + 1.
            idx = signals.level[self.groupby].get_loc(roi_label)
            signals.iloc[idx, frame_idxes.astype(int)] = 1.
            
        return signals
        
        
        



"""Convenience functions for experiment classes"""

import os
from copy import deepcopy
import json

import numpy as np
import h5py
from scipy.interpolate import interp1d
from scipy.signal import gaussian, convolve
from sima.ROI import ROIList

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, bytes):
            return obj.decode()

        elif isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):

            return int(obj)

        elif isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)
        
        elif isinstance(obj, (np.complex_, np.complex64, np.complex128)):
            return {'real': obj.real, 'imag': obj.imag}
        
        elif isinstance(obj, (np.bool_)):
            return bool(obj)

        elif isinstance(obj, (np.void)): 
            return None

        return json.JSONEncoder.default(self, obj)


def discretize_behavior(beh_dict, sampling_interval=0.1, discard_pre=0,
                        discard_post=np.inf, sigma=0.1):
    """Converts format of behavior data dictionary from continuous to discrete
    time. Intervals become boolean variables, and continuous variables are
    resampled at regular intervals.

    If there are additional variables to check for and retain in the
    behavior data, they should be added to this function.

    Parameters
    ----------
    sampling_interval : float, optional
        Sampling rate for discretizing time. Variables expressed as
        intervals (i.e. vectors of start and stop times) are converted to
        binary vectors that are True for frames inside the intervals.
        Continuous variables are resampled at the corresponding time points.
    discard_pre : {'first_lap', float}, optional
        If 'first_lap', truncate data before the first complete lap. If
        float, truncate data occuring before this time, in seconds. Default
        behavior returns all data points
    discard_post : float, optional
        Truncate data occuring after this time, in seconds. Default behavior
        returns all data points
    sigma : float, optional
        Standard deviation of gaussian smoothing kernel for smoothing
        velocity variable, in seconds. Defaults to 0.1

    Returns
    -------
    out : dict
        Behavior data dictionary with variables expressed in discrete time
        bins
    """

    # ---------------------------- PRELIMIANRIES ---------------------------- #

    total_duration = beh_dict['recordingDuration']
    out_dict = deepcopy(beh_dict)

    if discard_pre == 'first_lap':
        # the first lap never strictly starts at zero, so we have to build in
        # some tolerance
        try:
            FIRST_LAP_TOLERANCE = 0.001
            if beh_dict['treadmillPosition'][0, 1] > FIRST_LAP_TOLERANCE:
                discard_pre = beh_dict['lap'][0] / sampling_interval
            else:
                discard_pre = 0
        except KeyError:
            print("No lap information, continuing without!")

    # convert trim times to trim bins for slicing
    start = int(discard_pre / sampling_interval)
    stop = int(np.min([total_duration, discard_post]) / sampling_interval)
    out_dict['recordingDuration'] = (stop - start) * sampling_interval

    # ------------------------ DISCRETIZE VARIABLES ------------------------- #

    # convert interval variables
    out_dict.update(
        {k: interval_to_boolean(
                v, sampling_interval, total_duration
            )[start:stop]
         for k, v in out_dict.items() if is_interval(k, v)})

    # resample position
    out_dict.update({'treadmillPosition': continuous_to_discrete(
            out_dict['treadmillPosition'], sampling_interval, total_duration
        )[start:stop]})

    # ------------------------------ VELOCITY ------------------------------- #

    try:
        # newer behavior files should record dy explicitly
        dt, dy = out_dict['__position_updates'][:, 0:2].T

        # convert rotary encoder ticks to cm
        dy /= (out_dict['__trial_info']['position_scale'] * 10)

        # calculate velocity from dy/dt
        out_dict['velocity'] = calculate_velocity_from_updates(
            dy, dt, sampling_interval, total_duration, sigma=sigma)[start:stop]

    except KeyError:
        # otherwise calculate from the derivative of position
        out_dict['velocity'] = calculate_velocity(
            out_dict['treadmillPosition'], sampling_interval,
            out_dict['trackLength'], sigma=sigma)[start:stop]

    # ----------------------- MAKE NEW LAP VARIABLES ------------------------ #

    try:
        # find bin of each lap start (round up!)
        lap = [int(s - start) for s in
               np.ceil(beh_dict['lap'] / sampling_interval)
               if ((s >= start) and (s <= stop))]
        lap = np.unique([0] + lap)

        # create additional variable that gives the lap number in each
        # discrete bin
        lap_length = np.diff(np.append(lap, stop - start)).astype('int')
        lap_bin = np.concatenate(
            [[l] * n_bins for l, n_bins in enumerate(lap_length)])

        out_dict['lap'] = lap
        out_dict['lap_bin'] = lap_bin
    except KeyError:
        print("No lap information, continuing without!")

    # ------------------------------ CLEAN UP ------------------------------- #

    # log parameters
    out_dict.update({'sampling_interval': sampling_interval,
                     'discard_pre': discard_pre,
                     'discard_post': discard_post})

    # remove some extraneous information
    try:
        out_dict['json'] = out_dict['__trial_info']
    except KeyError:
        pass

    for k in list(out_dict.keys()):
        if '__' in k:
            out_dict.pop(k)

    return out_dict


def is_interval(key, val):
    """Helper function for finding interval variables from the standard behavior
    dictionary output"""

    if key in ['__position_y', 'treadmillPosition']:
        return False
    try:
        if val.shape[1] == 2:
            # intervals will have strictly two columns
            return True
        else:
            False
    except (AttributeError, IndexError):
        # either variable was not an array or did not have 2 dimensions
        return False


def interval_to_boolean(val, sampling_interval, total_duration):
    """Convert an interval variable to boolean vector

    Parameters
    ----------
    val : array of shape (n_samples, 2)
        Array of intervals. Each row is an event, where the first column gives
        the start time and the second column gives the stop time
    sampling_interval : float
        Width of each discrete time bin
    total_duration : int
        Sets the total length of the output vector, which will have length
        int(total_duration / sampling_interval)

    Returns
    -------
    out : array of length int(total_duration / sampling_interval)
        Boolean version of interval variable
    """
    out = np.zeros(int(total_duration / sampling_interval), 'bool')

    for i, (start, stop) in enumerate(val):

        # Is this still necessary if nans are fixed in tdmlPickler?
        if np.isnan(start):
            start = 0 if i == 0 else val[i-1, 0]
        if np.isnan(stop):
            try:
                stop = val[np.isfinite(val[:, 1])][i, 1]
            except IndexError:
                stop = len(out)

        start_frame = int(start / sampling_interval)
        stop_frame = int(stop / sampling_interval) + 1
        out[start_frame:stop_frame] = True

    return out


def continuous_to_discrete(val, sampling_interval, total_duration):
    """Resample a continuous variable (sampled at possibly variable intervals)
    to one sampled at regular, discrete time bins, via interpolation.

    Parameters
    ----------
    val : array of shape (n_samples, 2)
        Array of samples. Each row is a sample, where the first column gives
        the timestamp of the sample and the second column gives the value of
        the sample
    sampling_interval : float
        Width of each discrete time bin
    total_duration : int
        Sets the total length of the output vector, which will have length
        int(total_duration / sampling_interval)

    Returns
    -------
    out : array of length int(total_duration / sampling_interval)
        Resampled version of continuous variable
    """

    interp_func = interp1d(val[:, 0], val[:, 1], kind='zero',
                           bounds_error=False, fill_value=val[-1, 1])

    return interp_func(np.arange(0, total_duration, sampling_interval))


def calculate_velocity_from_updates(dy, dt, sampling_interval, total_duration,
                                    sigma=0.1):
    """Calculate animal velocity from dy/dt, as recorded from the rotary
    encoder '__position_updates' (for newer behavior files).

    Parameters
    ----------
    dy : array (n_samples,)
        Array of position updates, in centimeters
    dt : array (n_samples,)
        Array of time stamps for position updates, in seconds
    sampling_interval : float
        Time length of each bin in velocity trace
    total_duration : float
        Total length of trace, in seconds
    sigma : float, optional
        Standard deviation of the gaussian smoothing kernel to apply to the
        velocity trace, in seconds. Defaults to 0.1.

    Returns
    -------
    velocity : array (n_samples,)
    """

    bins = np.arange(0, total_duration + sampling_interval, sampling_interval)
    bin_idx = np.digitize(dt, bins=bins)

    # find sum displacements in each time bin
    dy_sums = [np.sum(dy[bin_idx == n]) for n in range(1, len(bins))]
    velocity = np.asarray(dy_sums) / sampling_interval  # convert to cm/sec

    # smooth
    sigma = int(np.ceil(sigma / sampling_interval))
    kernel = gaussian(8 * sigma, sigma)
    kernel /= kernel.sum()

    return convolve(velocity, kernel, mode='same')


def calculate_velocity(pos, sampling_interval, track_length,
                       sigma=0.1, max_discontinuity=0.2):
    """Calculate animal velocity from the change in position values across
    samples. Use this only if '__position_updates' is not available in the
    behavior dictionary.

    Parameters
    ----------
    pos : array (n_samples,)
        Array of position samples, scaled to range [0, 1]
    sampling_interval : float
        Time length of position bin
    track_length : float
        Length of track in millimeters
    sigma : float, optional
        Standard deviation of the gaussian smoothing kernel to apply to the
        velocity trace, in seconds. Defaults to 0.1.
    max_discontinuity : float (0, 1), optional
        Max allowable discontinuity in position from sample t to t + 1. If the
        position change is greater than this value, we assume a position reset
        occured (e.g. at the end of a lap, or ITI). The value of the violating
        bin is changed to match that of the preceding bin (i.e., assume a
        constant velocity from the prior sample during the reset). Defaults to
        0.2 (i.e. a fifth of the track length)

    Returns
    -------
    velocity : array (n_samples,)
    """

    delta_pos = np.concatenate([[0], np.diff(pos)])
    discont_idx = np.where(np.abs(delta_pos) > max_discontinuity)[0]

    # TODO : should we just interpolate?
    delta_pos[discont_idx] = delta_pos[discont_idx - 1]
    delta_pos *= (track_length / 10)  # change units to cm
    velocity = delta_pos / sampling_interval  # cm/sec

    # convert sigma to frames
    sigma = np.round(sigma / sampling_interval).astype('int')
    kernel = gaussian(sigma * 8, sigma)
    kernel /= kernel.sum()

    return convolve(velocity, kernel, mode='same')


def is_bruker(h5_path):
    try:
        get_prairie_xml_path(h5_path)
        return True
    except Exception:
        return False


def get_prairie_xml_path(h5_path):
    try:
        xml_name = [s.replace('.env', '.xml')
                    for s in os.listdir(os.path.split(h5_path)[0])
                    if s.endswith('.env')][0]
    except IndexError:
        xml_names = [s for s in os.listdir(os.path.split(h5_path)[0])
                     if s.endswith('.xml')]
        assert len(xml_names) == 1
        xml_name = xml_names[0] 

    xml_path = os.path.join(os.path.split(h5_path)[0], xml_name)
    if os.path.exists(xml_path):
        return xml_path
    else:
        raise Exception(
            "Could not locate Prairie xml/env files in h5 directory")


def get_prairie_frame_period(xml_path, round_result=True):
    """Parse the PrairieView xml to return the corrected frame period

    Parameters
    ----------
    xml_path : str
        Path to xml file
    round_result : bool, optional
        Whether to round the result to 6 decimal places. Defaults to True.
    """

    # TODO this is copied from the old experiment class and can definitely be
    # cleaned up considerably

    from xml.etree import ElementTree

    n_seqs = 2
    seqs = []
    _iter = ElementTree.iterparse(xml_path)
    while len(seqs) < n_seqs:
        try:
            _, elem = next(_iter)
        except StopIteration:
            break
        if elem.tag == 'Sequence':
            seqs.append(elem)

    if len(seqs) > 1:
        times = [float(s.find('Frame').get('absoluteTime'))
                 for s in seqs]
    else:
        times = [float(frame.get('absoluteTime'))
                 for frame in seqs[0].findall('Frame')[:n_seqs]]

    frame_period = np.diff(times).mean()
    if seqs[0].get('bidirectionalZ', 'False') == 'True':
        frame_period *= 2.0

    if round_result:
        return np.around(frame_period, 6)

    return frame_period


def is_scanimage(h5_path):
    raise NotImplementedError


def get_scanimage_frame_period():
    raise NotImplementedError


def is_femtonics(h5_path):
    raise NotImplementedError


def get_femtonics_frame_period():
    raise NotImplementedError


def update_h5_metadata(h5_path):
    """doc string"""

    h5 = h5py.File(h5_path, 'r+', libver='latest')

    if 'frame_period' not in h5['/imaging'].attrs.keys():

        if is_bruker(h5_path):
            xml_path = get_prairie_xml_path(h5_path)
            h5['/imaging'].attrs['frame_period'] = \
                get_prairie_frame_period(xml_path)
            h5['/imaging'].attrs['imaging_system'] = 'bruker'
        elif is_scanimage(h5_path):
            h5['/imaging'].attrs['frame_period'] = \
                get_scanimage_frame_period()
            h5['/imaging'].attrs['imaging_system'] = 'scanimage'
        elif is_femtonics(h5_path):
            h5['/imaging'].attrs['frame_period'] = \
                get_femtonics_frame_period()
            h5['/imaging'].attrs['imaging_system'] = 'femtonics'

    metadata = {k: v for k, v in h5['/imaging'].attrs.items()}
    h5.close()
    return metadata


def split_signals(expt, new_label, new_label_rois, current_label=None,
                  inverse=False, signal_type='*', channel='*', dry_run=False,
                  overwrite=False):
    """Split off a subset of ROIs to a new label, together with their signals.

    Parameters
    ----------
    expt : Experiment
        Experimet to get signals and rois from
    new_label : str
        New label to put new signals and ROIs in
    new_label_rois : array-like of ROI labels
        ROIs to move to new label. Only ROIs found within the current
        experiment will be included
    current_label : str
        Current label where the ROIs to be moved reside
    inverse : bool, optional
        Move all ROIs *not* in `new_label_rois` to new_label.
        Defaults to False.
    signal_type : str, optional
        The type of signal ({'raw', 'dfof', etc.}) to move. Defaults to all.
    channel : str, optional
        The channel ({'Ch1', 'Ch2', etc.}) to read signals from.
        Default to all.
    dry_run : bool, optional
        Show actions to be taken, without writing anything. Defaults to False.
    overwrite : bool, optional
        Overwrite key `new_label`, if it already exists in the signals file.
        Defaults to False.
    """
    print(f'In {expt.sima_path}:')
    rois = expt.roi_list(label=current_label)

    if inverse:
        new_rois = ROIList(
            [r for r in rois if r.label not in set(new_label_rois)])
    else:
        new_rois = ROIList([r for r in rois if r.label in set(new_label_rois)])

    roi_path = os.path.join(expt.sima_path, 'rois.pkl')

    with expt.signals_file() as signals_file:
        keys = signals_file.keys()

    for key in keys:
        _, ch_key, label_key, type_key = key.split('/')
        if (ch_key == channel or channel == '*') \
                and (type_key == signal_type or signal_type == '*') \
                and (label_key == current_label):

            signals = expt.signals(channel=ch_key, signal_type=type_key,
                                   label=current_label)

            if inverse:
                new_signals = signals.loc[
                    signals.index.difference(new_label_rois)]
            else:
                new_signals = signals.loc[
                    signals.index.intersection(new_label_rois)]

            # TODO: Handle this in a better way
            new_key = f'/{ch_key}/{new_label}/{type_key}'

            print(f"Splitting off {len(new_signals)} ROIs "
                  + f"from {key} -> {new_key}")
            with expt.signals_file(mode='a') as write_file:
                key_exists = new_key in write_file
                assert not key_exists or overwrite, \
                    f"Key {new_key} already exists in signal file. Set " \
                    + "overwrite=True to recalculate"
                if not dry_run:
                    write_file.put(new_key, new_signals)

    print(f"Splitting off {len(new_rois)} ROIs from "
          + f"{current_label} -> {new_label}")
    if not dry_run:
        new_rois.save(roi_path, label=new_label)



import pandas as pd 

from lab3.core.classes import Group
from . import IntervalEvent, EventGroup, EventsFile
from .group import Group2DataFrameMixin, ImagingEventGroupGroup       

class Transient(IntervalEvent):
    constructor_ids = ['onset', 'offset', 'sigma', 'duration', 
                       'amplitude', 'amplitude_index']
    #def __init__(self, onset, offset, sigma, duration, amplitude, amplitude_index):
    #    self.onset = onset
    #    self.offset = offset
    #    self.sigma = sigma
    #    self.duration = duration
    #    self.amplitude = amplitude
    #    self.amplitude_index = amplitude_index
    
class SingleCellTransients(EventGroup):
    item_class = Transient

class ExperimentTransientGroup(ImagingEventGroupGroup):
    item_class = SingleCellTransients

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            for single_cell_transients in self:
                if single_cell_transients.name == key:
                    return single_cell_transients
                
    @classmethod
    def from_pandas(cls, dataframe, trigger=None, **kwargs):
        return cls([SingleCellTransients.from_pandas(df, name=roi_label, trigger=trigger) 
                    for roi_label, df in dataframe.groupby("roi_label")])





import pandas as pd
import numpy as np
import copy
import scipy.signal as ss

from lab3.core import Item, Group

class EventsFile(pd.HDFStore):
    pass

class Event(Item):
    name = 'Event'
    constructor_ids = ['onset', 'offset']

    def __init__(self, onset=None, offset=None, trigger=None, **kwargs):
        self.onset = onset 
        self.offset = offset
        self.trigger_key = trigger

        super().__init__(**kwargs)

    def label(self):
        return str((self.onset, self.offset))

    def validate(self):
        """Check that this event is valid. Useful if writing to events.h5
        More important for subclasses of Event
        """
        assert 'onset' in self
        # assert self.duration > 0
        self.point_estimate()
        
    @property
    def duration(self):
        try:
            return self.offset - self.onset
        except (AttributeError, TypeError):
            return 0
    
    @property
    def trigger(self):
        """Localize the event to a single point in time.
        """
        try:
            return self.__dict__[self.trigger_key]
        except:
            return self.onset

    def to_frame(self):
        """Localize the event to a frame.
        This needs to be overridden for non-imaging events
        """
        return self.trigger 

    @classmethod
    def from_pandas(cls, series, **kwargs):
        return cls(**series, **kwargs)

    def to_pandas(self):
        return pd.Series(self.constructors(), name=self.name)

    def __getitem__(self, key):
        return self.__dict__[key]
    
    def __contains__(self, time):
        return self.onset <= time < self.offset 

class TriggerEvent(Event):
    constructor_ids = ['onset']

    def __init__(self, onset):
        super().__init__(onset=onset, offset=onset)

    def label(self):
        return self.onset

class OnsetEvent(TriggerEvent):
    pass

class OffsetEvent(TriggerEvent):
    @property
    def trigger(self):
        return self.offset

class IntervalEvent(Event):
    pass

def bool2events(array, event_class=Event):
    events = []
    
    i = 0
    in_event = False
    onset = None
    offset = None
    while i < len(array)-1:
        if array[i] and not in_event:
            onset = i
            in_event = True
        elif not array[i] and in_event:
            offset = i
            in_event = False            
            events.append(event_class(onset=onset, offset=i))
            onset = None
            offset = None
        i += 1
    if in_event:
        events.append(event_class(onset=onset, offset=i))
    return events
        
    
    
    
        
    



"""Simple tools for data sharing"""

# TODO: it would be nice to have experiment classes that can read from these
# files so that the same analysis can be done without the database
# Right now we just handle simple exports

import os

import pandas as pd


class H5DataDumper:

    """This object takes experiments and dumps them to HDFStores.

    For now this only works on ImagingExperiments, but we'll make it more
    flexible in the future with options about which data/metadata to store.

    Parameters
    ----------
    savepath : str
        Where to save the h5 stores
    binarize : bool, optional
        If dumping spikes, whether tto binarize signals first.
    """

    def __init__(self, savepath, binarize=True):

        self.savepath = savepath
        self.binarize = binarize

    @property
    def savepath(self):
        return self._savepath

    @savepath.setter
    def savepath(self, p):
        if not os.path.exists(p):
            os.mkdir(p)
        self._savepath = p

    def _dump_imaging_data(self, expt, store):

        for key in expt.signals_file().keys():
            # what happens if there are no signals? need error handling here
            # for non-imaging experiments
            _, channel, label, signal_type = key.split(os.sep)

            # we retrieve the signals via the method so that, if we're dealing
            # with an imaging experiment, it gets trimmed to match the behavior
            # under the hood
            signal = expt.signals(channel=channel, label=label,
                                  signal_type=signal_type)
            if (signal_type == 'spikes') & self.binarize:
                signal = signal.gt(0)

            store.put(key, signal)

    def _dump_behavior_data(self, expt, store):

        behavior = {}
        for key, value in expt.format_behavior_data().items():
            try:
                # TODO - handle max frame when not ImagingMixin
                if len(value) == len(expt.imaging_times()):
                    behavior[key] = value
            except Exception:  # TypeError?
                pass

        behavior = pd.DataFrame(behavior)
        store.put('behavior', behavior)

    def _dump_metadata(self, expt, store):
        # TODO
        pass

    def _datadump(self, expt, store):
        self._dump_imaging_data(expt, store)
        self._dump_behavior_data(expt, store)
        self._dump_metadata(expt, store)

    def dump(self, expt, overwrite=False):
        """Given an experiment object, create the h5 store"""
        h5_path = os.path.join(self.savepath, f'{expt.trial_id}.h5')
        if os.path.isfile(h5_path):
            if overwrite:
                os.remove(h5_path)
            else:
                print(f"'{h5_path}' exists and overwrite=False, skipping...")
                return

        with pd.HDFStore(h5_path) as store:
            self._datadump(expt, store)


class VRDataDumper(H5DataDumper):

    def _dump_behavior_data(self, expt, store):

        behavior = {
            'velocity': expt.velocity(),
            'position': expt.discrete_position(),
            'iti': expt.iti(),
            'valid': expt.valid_samples(),
            'reward': expt.format_behavior_data()['reward'],
            'licking': expt.format_behavior_data()['licking'],
            'lap': expt.format_behavior_data()['lap_bin'],
        }
        behavior = pd.DataFrame(behavior)
        store.put('behavior', behavior)

    def _dump_masks(self, expt, store):
        masks = pd.DataFrame(
            [{'roi_label': roi.label, 'roi_mask': roi.mask[0]}
             for roi in expt.roi_list()]).set_index('roi_label')
        store.put('roi_masks', masks)

    def _dump_timeavg(self, expt, store):
        ds = expt.suite2p_imaging_dataset
        time_avg = ds.ops_list[0]['meanImg']
        store.put('time_avg', pd.DataFrame(time_avg))

    def _datadump(self, expt, store):
        super()._datadump(expt, store)
        self._dump_masks(expt, store)
        self._dump_timeavg(expt, store)

    def dump(self, expt, overwrite=False):
        from lab3.experiment.virtual import VirtualImagingExperiment
        assert isinstance(expt, VirtualImagingExperiment)
        super().dump(expt, overwrite=overwrite)



"""Methods for spike inferencing. Implement custom strategies by subclassing
SpikesStrategy and filling in the details."""

from abc import abstractmethod
import warnings

import numpy as np
import pandas as pd
from multiprocessing import Pool
from oasis.functions import deconvolve  # should we move this into the class?

from lab3.core import Automorphism

warnings.filterwarnings("ignore", category=FutureWarning)


class SpikesStrategy(Automorphism):

    """Base class for spike inferencing strategies. Derived strategies should
    implement the _calculate_baseline and _do_inference methods."""

    name = "spikes"

    def apply_to(self, experiment, channel='Ch2', label=None, from_type='dfof',
                 **kwargs):
        """Apply spike inferencing analysis to an experiment object.

        Parameters
        ----------
        experiment : lab3.experiment.base.ImagingExperiment
            Instance of ImagingExperiment to analyze
        channel : str or int, optional
        label : str, optional
        from_type : str {'dfof', 'raw'} or custom, optional

        Additional keyword arguments are passed directly to `calculate`.
        """
        sig_dict = self._load_signals(experiment, channel=channel, label=label,
                                      from_type='dfof')
        return self.calculate(**sig_dict, **kwargs)

    def calculate_baseline(self, signals):
        try:
            return self.baseline
        except AttributeError:
            self.baseline = self._calculate_baseline(signals)
            return self.baseline

    def calculate(self, signals):
        """Run spike inferencing"""

        try:
            del self.baseline
        except AttributeError:
            pass
        baseline = self.calculate_baseline(signals)
        return self._do_inference(signals - baseline)

    # ----------------- Implement these methods to subclass ----------------- #

    @abstractmethod
    def _do_inference(self, signals):
        """Implement the spike inferencing algorithm here. It should return
        a dataframe the same size as signals, preserving the index"""

    # ------------------------------- Optional ------------------------------ #

    def _calculate_baseline(self, signals):
        """Estimate baseline for dFOF calculation. It should return a dataframe
        the same size as signals, preserving the index"""
        return np.zeros(signals.shape)

    def _load_signals(self, experiment, channel, label, from_type):
        """By default only a single signal dataframe is passed to `calculate`
        when using `apply`. If you override `calculate` to take additional
        arguments in your custom strategy, you should load and add them to the
        dictionary here, which will be unpacked as arguments to `calculate`.
        """
        sig_dict = {
            'signals': experiment.signals(signal_type=from_type,
                                          channel=channel, label=label,
                                          max_frame=None)
        }
        return sig_dict


class MADMixin:

    """Mixin that adds an additional spike filtering step following inference,
    by discarding spikes with amplitude less than some threshold number of
    median absolute deviations from the original trace. Since we override
    the calculate method, this should be positioned first in the inheritance
    list when subclassing.

    Derived classes should store an n_mad attribute, for calculating the
    thresholds, and expose this as a parameter to the user in the init
    method.
    """

    # TODO : should we just implement this using the min_spike in oasisAR1?

    def _mad_filter_spikes(self, signals, spikes):
        """Do MAD filtering"""

        mad = signals.sub(signals.median(axis=1), axis=0).abs().median(axis=1)
        self.thresholds = mad * self.n_mad
        spikes[spikes.le(self.thresholds, axis=0)] = 0
        return spikes

    def calculate(self, signals):
        """Calculates spikes and applies MAD filtering, if the n_mad attribute
        exists"""
        spikes = super().calculate(signals)

        try:
            return self._mad_filter_spikes(signals - self.baseline, spikes)
        except AttributeError:
            return spikes


def _deconvolve(inputs):
    """Must be a top level function for multiprocessing"""
    label, signal, g, kws = inputs
    return label, deconvolve(np.nan_to_num(signal), g=g, **kws)[1]


def _iter_signals(signals, g, kws):
    for label, signal in signals.iterrows():
        yield (label, signal, g, kws)


class OasisAR1Spikes(MADMixin, SpikesStrategy):

    """Infer spikes using the OASIS method, AR1 model variaton.

    TODO framerate should be in the metadata of the signals dataframe, so we
    don't have to pass it explicitly

    Parameters
    ----------
    tau : float, optional
        Decay time constant of the indicator, in seconds. Defaults to 0.7,
        which is a reasonable number for GCaMP6f.
    fs : float, optional
        Frame rate, in Hz. Defaults to 10.
    n_mad : int, optional
        Filter spikes following inference, discarding any events with an
        amplitude less than n_mad * median absolute deviation of the original
        signal trace (e.g., 3 n_mad corresponds roughly to discarding spikes
        with an amplitude less than 2 standard deviations of the original
        trace). Defaults to 0 (no filtering).
    n_processes : int, optional
    oasis_kws : dict, optional
        description

    Returns
    -------
    spikes : pd.DataFrame

    Attributes
    ----------
    baseline : pd.DataFrame
    thresholds : pd.Series
    """

    def __init__(self, tau=0.7, fs=10, n_mad=3, n_processes=1, oasis_kws={}):
        self.tau = tau
        self.fs = fs
        self.n_mad = n_mad
        self.oasis_kws = oasis_kws
        self.n_processes = n_processes

    @property
    def g(self):
        """Calculates g for the AR1 model"""
        return np.exp(-1 / (self.tau * self.fs))

    def _do_inference(self, signals):
        spikes = pd.DataFrame(np.zeros(signals.shape), index=signals.index)

        _signals = _iter_signals(signals, [self.g], self.oasis_kws)

        if self.n_processes > 1:
            p = Pool(self.n_processes)
            for l, s in p.map(_deconvolve, _signals):
                spikes.loc[l] = s

            p.close()
            p.join()

        else:
            for l, s in map(_deconvolve, _signals):
                spikes.loc[l] = s

        return spikes


class OasisAR2Spikes(MADMixin, SpikesStrategy):

    """Infer spikes using the OASIS method, AR1 model variaton.

    Parameters
    ----------
    tau_d, tau_r : float, optional
        Decay and rise time constant of the indicator, in seconds. Defaults to
        0.7 and 0.065 respectively, which are reasonable numbers for GCaMP6f.
    fs : float, optional
        Frame rate, in Hz. Defaults to 10.
    n_mad : int, optional
        description
    oasis_kws : dict, optional
        description

    Returns
    -------
    spikes : pd.DataFrame

    Attributes
    ----------
    baseline : pd.DataFrame
    thresholds : pd.Series
    """

    # TODO : Does this need to be a separate class? On the plus side it avoids
    # overloading tau? Though we could choose AR1 vs AR2 by whether tau_r
    # is passed?

    def __init__(self, tau_r=0.065, tau_d=0.7, fs=10, n_mad=3, oasis_kws={}):
        self.tau_r = tau_r
        self.tau_d = tau_d
        self.fs = fs
        self.n_mad = n_mad
        self.oasis_kws = oasis_kws

    @property
    def g(self):
        """Calculate g for the AR2 model"""
        term1 = np.exp(-1 / (self.tau_d * self.fs))
        term2 = np.exp(-1 / (self.tau_r * self.fs))
        return (term1 + term2, -1 * term1 * term2)

    def _do_inference(self, signals):

        from oasis.functions import deconvolve

        spikes = pd.DataFrame(np.zeros(signals.shape),
                              index=signals.index)
        for roi, signal in signals.iterrows():
            spikes.loc[roi] = deconvolve(signal.values, g=self.g,
                                         **self.oasis_kws)[1]

        return spikes



import os
import warnings

# Currently we can't limit multithreading directly in Suite2p, so we must set
# these environment variables prior to numpy/suite2p imports
N_PROCESSES = str(8)
os.environ["VECLIB_MAXIMUM_THREADS"] = N_PROCESSES
os.environ["NUMEXPR_NUM_THREADS"] = N_PROCESSES
os.environ["NUMBA_NUM_THREADS"] = N_PROCESSES
os.environ["NUMBA_DEFAULT_NUM_THREADS"] = N_PROCESSES
os.environ["OMP_NUM_THREADS"] = N_PROCESSES
os.environ["OPENBLAS_NUM_THREADS"] = N_PROCESSES
os.environ["MKL_NUM_THREADS"] = N_PROCESSES

OPS_0_6_16 = {

    # file paths
    'look_one_level_down': False,
    'fast_disk': [],
    'delete_bin': False,  # whether to delete binary file after processing
    'mesoscan': False,
    'h5py': [],
    'h5py_key': 'data',
    'save_path0': [],
    'save_folder': [],
    'subfolders': [],
    'bruker': False,

    # main settings
    'nplanes': 1,  # each tiff has these many planes in sequence
    'nchannels': 1,  # each tiff has these many channels per plane
    # this channel is used to extract functional ROIs (1-based)
    'functional_chan': 1,
    # calcium time constant -- this is the main parameter for deconvolution
    'tau':  0.7,
    # sampling rate -- for multi-plane imaging, this is the sampling rate per
    # complete stack/volume
    'fs': 10.,
    'force_sktiff': False,

    # output settings
    'preclassify': 0.,
    'save_mat': False,
    # combine multiple planes into a single result /single canvas for GUI
    # don't change this, or our import step will not work properly
    'combined': True,
    # um/pixels in X / um/pixels in Y (for correct aspect ratio in GUI)
    'aspect': 1.0,

    # bidirectional phase offset
    'do_bidiphase': False,
    'bidiphase': 0,

    # registration settings
    # whether to register data (2 forces re-registration)
    'do_registration': 1,
    'keep_movie_raw': False,
    # subsampled frames for finding reference image
    'nimg_init': 300,
    # number of frames per batch
    'batch_size': 2000,
    # max allowed registration shift, as a fraction of
    # frame max(width and height)
    'maxregshift': 0.1,
    # when multi-channel, you can align by non-functional channel (1-based)
    'align_by_chan': 1,
    'reg_tif': False,  # whether to save registered tiffs
    'reg_tif_chan2': False,  # whether to save channel 2 registered tiffs
    'subpixel': 10,  # precision of subpixel registration (1/subpixel steps)
    # ~1 good for 2P recordings, recommend >5 for 1P recordings
    'smooth_sigma': 1.15,
    # this parameter determines which frames to exclude when determining
    # cropping - set it smaller to exclude more frames
    'th_badframes': 1.0,
    'pad_fft': False,

    # non rigid registration settings
    'nonrigid': True,  # whether to use nonrigid registration
    # block size to register (** keep this a multiple of 2 **)
    'block_size': [128, 128],
    # if any nonrigid block is below this threshold, it gets smoothed until
    # above this threshold. 1.0 results in no smoothing
    'snr_thresh': 1.2,
    # maximum pixel shift allowed for nonrigid, relative to rigid
    'maxregshiftNR': 5,

    # 1P settings
    '1Preg': False,
    'spatial_hp': 25,
    'pre_smooth': 2,
    'spatial_taper': 50,

    # cell detection settings
    'roidetect': True,  # whether or not to run ROI extraction
    'sparse_mode': False,  # whether or not to run sparse_mode
    # use diameter for filtering and extracting (ignored for sparse_mode=True)
    'diameter': 12,
    # use spatial scale for sparse_mode (ignored for sparse_mode=False)
    # 0: multi-scale; 1: 6 pixels, 2: 12 pixels, 3: 24 pixels, 4: 48 pixels
    'spatial_scale': 1,
    # whether or not to keep ROIs fully connected (set to 0 for dendrites)
    'connected': True,
    # max number of binned frames for cell detection
    'nbinned': 10000,
    # maximum number of iterations to do cell detection
    'max_iterations': 20,
    # adjust the automatically determined threshold by this scalar multiplier
    'threshold_scaling': 1.0,
    # cells with more overlap than this get removed during triage, before
    # refinement
    'max_overlap': 0.75,
    # running mean subtraction with window of size 'high_pass'
    # (use low values for 1P)
    'high_pass': 100,

    # ROI extraction parameters
    # number of pixels to keep between ROI and neuropil donut
    'inner_neuropil_radius': 2,
    # minimum number of pixels in the neuropil
    'min_neuropil_pixels': 350,
    # pixels that are overlapping are thrown out (False)
    # or added to both ROIs (True)
    'allow_overlap': False,

    # channel 2 detection settings (stat[n]['chan2'], stat[n]['not_chan2'])
    'chan2_thres': 0.65,  # minimum for detection of brightness on channel 2

    # deconvolution settings
    'baseline': 'maximin',  # baselining mode (can also choose 'prctile')
    'win_baseline': 60.,  # window for maximin
    'sig_baseline': 10.,  # smoothing constant for gaussian filter
    'prctile_baseline': 8.,  # optional (whether to use a percentile baseline)
    'neucoeff': .7,  # neuropil coefficient
}

DB = {

    # these fields are generated by Suite2pImagingDataset objects automatically
    'h5py': [],
    'h5py_key': 'data',
    'look_one_level_down': False,
    'data_path': None,
    'subfolders': [],
    'fast_disk': None,
}

OPS_0_5_5 = {}


def get_ops(suite2p_version):
    """Returns the default options dictionary for Suite2p, specific to the
    suite2p version installed.

    Parameters
    ----------
    suite2p_version : str
        Suite2p version number. They don't include this under
        suite2p.__version__ for whatever reason, but you can access it in
        python via pkg_resources.get_distribution(suite2p)

    Returns
    -------
    ops : dict
        Default options dictionary
    """
    if suite2p_version == '0.6.16':
        return OPS_0_6_16
    elif suite2p_version == '0.5.5':
        return OPS_0_5_5
    elif suite2p_version == '0.7.5':
        return OPS_0_6_16
    else:
        warnings.warn("Untested suite2p version! Using 0.6.16 ops dictionary")
        return OPS_0_6_16



"""Base signal classes"""

import inspect
from pandas import Series, DataFrame, HDFStore


class SignalFile(HDFStore):

    pass
    # def __init__(self, *args, **kwargs):
    #     super().__init__(self, *args, **kwargs)


class SignalFrame(DataFrame):

    """doc string"""

    def __init__(self, signals, metadata={}, **kwargs):
        super().__init__(signals)
        self.metadata = metadata
    #     self.cls = type(signals[0])
    #
    # def indexed_signal(self, index):
    #     return self.cls(self.iloc[index], **self.metadata.iloc[index])


class CalciumSignals(SignalFrame):
    pass


class DFOFSignals(SignalFrame):
    pass


class LFPSignals(SignalFrame):
    pass


class SpikeSignals(SignalFrame):
    pass


class Signal(Series):
    pass


class BaseSignalTransformer:

    """Base class for all signal transformation strategies. Adapted largely
    from sklearn.base.BaseEstimator"""

    @classmethod
    def _get_param_names(cls):
        init_signature = inspect.signature(cls.__init__)
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]

        return sorted([p.name for p in parameters])

    def get_params(self):
        out = dict()
        for key in self._get_param_names():
            try:
                value = getattr(self, key)
            except AttributeError:
                value = None
            out[key] = value
        return out

    def set_params(self):
        raise NotImplementedError

    def __repr__(self, line_max=80):
        """TODO : Make this pretty"""
        repr_ = type(self).__name__
        repr_ += '('
        base_length = len(repr_)
        curr_line_length = len(repr_)
        for key, value in self.get_params().items():
            param_string = key + '=' + str(value) + ', '
            if (curr_line_length + len(param_string)) <= line_max:
                repr_ += param_string
                curr_line_length += len(param_string)
            else:
                repr_ += '\n' + ' ' * base_length
                repr_ += param_string
                curr_line_length = base_length + len(param_string)
        repr_ = repr_[:-2] + ')'
        return repr_

    def __str__(self):
        return self.__repr__()

    def __get_state__(self):
        pass

    def __set_state__(self):
        pass



from __future__ import annotations
import pandas as pd 
import numpy as np
from functools import reduce
from typing import TYPE_CHECKING

from lab3.core.filters import TimeFilter, _BoundFilter
from lab3.event import SpatialEvent

if TYPE_CHECKING:
    from lab3.experiment import ImagingExperiment

class SpatialFilter(TimeFilter, SpatialEvent):
    """Select timepoints where the animal was in a spatial location
    
    SpatialFilters can be combined with the usual Filter grammar.
    """
    def bind_to(self, expt : ImagingExperiment):
        events = super().to_time_events(expt)
        idxes = [pd.RangeIndex(ev.onset, ev.offset) for ev in events]
        return self.bind(reduce(pd.Index.union, idxes))

class EventFilter(TimeFilter):
    """Selects activity inside events, specified with a dictionary.
    Example
    -------
    >>> event_spec = {
            'signal_type': 'behavior',
            'event_type': 'running',
            'as_time_events': False
            }
    >>> running_filter = EventFilter(event_spec)
    >>> is_running = running_filter.bind_to(expt)
    >>> is_not_running = ~is_running

    """
    constructor_ids = ['event_spec']
    def __init__(self, event_spec):
        self.event_spec = event_spec

    def bind_to(self, expt : ImagingExperiment):
        events = expt.events(**self.event_spec)
        idxes = [pd.RangeIndex(ev.onset, ev.offset) for ev in events]
        return self.bind(reduce(pd.Int64Index.union, idxes, pd.Int64Index([])))

class IsRunning(EventFilter):
    constructor_ids = ['threshold', 'debounce_window', 'as_time_events']
    def __init__(self, threshold=4, debounce_window=4, as_time_events=False):
        self.threshold = threshold
        self.debounce_window = debounce_window
        self.as_time_events = as_time_events
        
        event_spec = {
            'signal_type': 'behavior', 
            'event_type': 'running',
            'fmt': 'interval',
            'threshold': threshold,
            'debounce_window': debounce_window,
            'as_time_events': as_time_events
        }
        
        super().__init__(event_spec=event_spec)

class InReward(EventFilter):
    constructor_ids = []
    def __init__(self):
        event_spec = {
            'signal_type': 'behavior', 
            'event_type': 'reward',
            'fmt': 'interval',
            'as_time_events': False            
        }
        
        super().__init__(event_spec=event_spec)

class RealTimeEventFilter(EventFilter):
    constructor_ids = ['event_spec', 'pre', 'post']
    def __init__(self, event_spec, pre=0., post=0.):
        self.pre = pre
        self.post = post

        super().__init__(event_spec=event_spec)


    def bind_to(self, expt):

        events = expt.events(**self.event_spec)
        imaging_times = expt.imaging_times()

        idxes = []
        for ev in events:
            lower, = np.where(imaging_times >= ev.onset - self.pre)
            upper, = np.where(imaging_times <= ev.offset + self.post)
            idxes.append(pd.RangeIndex(lower[0], upper[-1]))

        return self.bind(reduce(pd.Int64Index.union, idxes, pd.Int64Index([])))



    
